/* simo.h

   originally written by Frederic Bois

   Copyright (c) 1993-2004.  Don Maszle, Frederic Bois.  All rights reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   -- Revisions -----
     Logfile:  %F%
    Revision:  %I%
        Date:  %G%
     Modtime:  %U%
      Author:  @a
   -- SCCS  ---------

   Header file for simo.c
*/

#ifndef _SIMO_H_
#define _SIMO_H_

/* ----------------------------------------------------------------------------
   Inclusions
*/

#include "yourcode.h"
#include "sim.h"


/* ----------------------------------------------------------------------------
   Prototypes
*/

void CloseMCFiles (PANALYSIS);
void SaveOutputs (PEXPERIMENT, PDOUBLE);
void NextOutputTime (PEXPERIMENT, PDOUBLE, PINT);
int OpenMCFiles (PANALYSIS panal);
void WriteMCHeader (PFILE, PANALYSIS);
void WriteMCOutput (PANALYSIS, PMCPREDOUT);
void WriteNormalOutput (PANALYSIS, PEXPERIMENT);
int  WriteOneMod (PVOID, PVOID);

#endif /* _SIMO_H_ */

/* End */

